/***********      .---.         .-"-.      *******************\
* -------- *     /   ._.       /  ` \     * ---------------- *
* Author's *     \_  (__\      \_v_/     * humus@rogers.com *
*   note   *     //   \\       //   \\     * ICQ #47010716    *
* -------- *    ((     ))     ((     ))    * ---------------- *
*          ****--""---""-------""---""--****                  ********\
* This file is a part of the work done by Humus. You are free to use  *
* the code in any way you like, modified, unmodified or copy'n'pasted *
* into your own work. However, I expect you to respect these points:  *
*  @ If you use this file and its contents unmodified, or use a major *
*    part of this file, please credit the author and leave this note. *
*  @ For use in anything commercial, please request my approval.      *
*  @ Share your work and ideas too as much as you can.                *
\*********************************************************************/

#include "App.h"
#include <time.h>

BaseApp *app = new App();

#define BATCH_INSTANCES 64

struct Vertex {
	Vertex(){}
	Vertex(const float x, const float y, const float z, const vec3 &n, const float i){
		pos.x = x;
		pos.y = y;
		pos.z = z;
		norm = n;
		id = i;
	}

	vec3 pos;
	vec3 norm;
	float id;
};

void App::resetCamera(){
	camPos = vec3(-10, 12, 0);
	wx = 0;
	wy = -PI / 2;
}

bool App::onKey(const uint key, const bool pressed){
	if (OpenGLApp::onKey(key, pressed)) return true;

	if (pressed){
		if (key == KEY_F5){
			freeFly->setChecked(!freeFly->isChecked());
		}
	}

	return false;
}

void App::controls(){
	if (freeFly->isChecked()){
		OpenGLApp::controls();
	}
}

bool App::init(){
	speed = 60;

	// Init GUI components
	int tab = configDialog->addTab("Domino");
	configDialog->addWidget(tab, freeFly = new CheckBox(0, 0, 200, 36, "Free fly", false));


	srand((uint) ::time(NULL));

	arrangeTiles();

	return true;
}

void App::exit(){

}

bool App::findCollision(const vec2 &pos){
	int count = tiles.getCount() - 3;

	for (int i = count; i >= 0; i--){
//	for (int i = 0; i < count; i++){
		vec2 dir = tiles[i].xy() - pos;
		if (dot(dir, dir) < 6 * 6) return true;
	}

	return false;
}

void App::arrangeTiles(){
	attractionPoint = vec2(0, 0);
	attractionTile = 0;
	tiles.setCount(0);


	vec2 pos(0, 0);
	float angle = 0;
	float fallTime = 0;
	
	int loops = 0;
	int failCount = 0;

	while (tiles.getCount() < 512 && loops < 4096){
		int startCount = tiles.getCount();
		bool restart = false;

		// First is always straight line, otherwise 40% line, 40% arc, 20% diamond
		int type;
		if (tiles.getCount()){
			type = rand() % 5;
		} else {
			type = 0;
		}

		if (type < 2){
			// Straight line
			int count = 2 + (rand() % 15);
			vec2 dir = 3.0f * vec2(cosf(angle), sinf(angle));

			for (int i = 0; i < count; i++){
				if (findCollision(pos)){
					restart = true;
					break;
				}

				tiles.add(vec4(pos, angle, fallTime));

				pos += dir;

				fallTime += 0.15f;
			}
		} else if (type < 4){
			// Arc
			float radius = 6.0f + float(rand()) / (0.05f * RAND_MAX);
			float angleStep = 3.0f / radius;

			float n = sqrtf(float(rand()) / RAND_MAX);
			int count = 2 + int(n * radius);

			vec2 dir = vec2(sinf(angle), -cosf(angle));

			vec2 center = pos;
			// Randomize left or right direction
			if (rand() & 1){
				 angleStep = -angleStep;
				 radius = -radius;
			}
			center -= radius * dir;

			
			for (int i = 0; i < count; i++){
				if (findCollision(pos)){
					restart = true;
					break;
				}

				tiles.add(vec4(pos, angle, fallTime));

				angle += angleStep;
				pos = center + radius * vec2(sinf(angle), -cosf(angle));

				fallTime += 0.15f;
			}
		} else {
			// Diamond
			int count = 5 + 2 * (rand() % 8);
			vec2 dir  = 3.0f * vec2(cosf(angle),  sinf(angle));
			vec2 sDir = 2.0f * vec2(sinf(angle), -cosf(angle));

			for (int i = 0; i < count; i++){
				float k;
				if (i < count / 2){
					k = float(i);
				} else {
					k = float(count - 1 - i);
				}
				vec2 p0 = pos + k * sDir;
				vec2 p1 = pos - k * sDir;

				if (findCollision(p0) || findCollision(p1)){
					restart = true;
					break;
				}

				tiles.add(vec4(p0, angle, fallTime));
				if (k > 0.5f){
					tiles.add(vec4(p1, angle, fallTime));
				}

				pos += dir;

				fallTime += 0.15f;
			}
		}

		// If we collided, then restart
		if (restart){
			failCount++;
			// Start back a good deal if we failed numerous times
			if (failCount > 32){
				startCount = max(startCount - 128, 0);
				failCount = 0;
			} else {
				startCount--;
			}
			pos = tiles[startCount].xy();
			angle = tiles[startCount].z;
			fallTime = tiles[startCount].w;

			tiles.setCount(startCount);
		}

		loops++;
	}
}


bool App::load(){
	if (!GLSL_supported){
		ErrorMsg("No GLSL support");
		return false;
	}

	// Shaders
	if ((floorShader = renderer->addShader("floor.shd")) == SHADER_NONE) return false;
	if ((tileShader  = renderer->addShader("tile.shd", DEFINE_STR(BATCH_INSTANCES))) == SHADER_NONE) return false;

	// Samplerstates
	if ((trilinear = renderer->addSamplerState(TRILINEAR, WRAP, WRAP, WRAP)) == SS_NONE) return false;
	if ((trilinearAniso = renderer->addSamplerState(TRILINEAR_ANISO, WRAP, WRAP, WRAP)) == SS_NONE) return false;


	// Textures
	if ((floorBase = renderer->addTexture  ("../Textures/block08d.dds",     true, trilinearAniso)) == TEXTURE_NONE) return false;
	if ((floorBump = renderer->addNormalMap("../Textures/block08dBump.dds", FORMAT_RGB565, true, trilinearAniso, 0.7f, 2.5f)) == TEXTURE_NONE) return false;

	// Generate a noise texture
#define SCALE 0.071622f

	Image img;
	if (!img.loadDDS("Noise.dds")){
		initNoise();

		ushort *dest = (ushort *) img.create(FORMAT_I16, 64, 64, 64, 1);

		for (int z = 0; z < 64; z++){
			for (int y = 0; y < 64; y++){
				for (int x = 0; x < 64; x++){
					float n = tileableTurbulence3(x * SCALE, y * SCALE, z * SCALE, 64 * SCALE, 64 * SCALE, 64 * SCALE, 8.0f);
					n = saturate(n * 1.5f + 0.5f);

					*dest++ = (ushort) (n * 65535.0f + 0.5f);
				}
			}
		}
		img.saveDDS("Noise.dds");
	}
	img.createMipMaps();
	if ((noise = renderer->addTexture(img, trilinear)) == TEXTURE_NONE) return false;



	// Generate BATCH_INSTANCES number of tiles
	Vertex vertices[6 * 4 * BATCH_INSTANCES];

	Vertex *dest = vertices;
	for (int i = 0; i < BATCH_INSTANCES; i++){
		float id = float(i);

		vec3 r( 1,  0,  0);
		vec3 l(-1,  0,  0);
		vec3 u( 0,  1,  0);
		vec3 d( 0, -1,  0);
		vec3 f( 0,  0,  1);
		vec3 b( 0,  0, -1);

		*dest++ = Vertex( 0,8,-2, r, id); *dest++ = Vertex( 0,8, 2, r, id); *dest++ = Vertex( 0,0, 2, r, id); *dest++ = Vertex( 0,0,-2, r, id);
		*dest++ = Vertex(-1,8, 2, l, id); *dest++ = Vertex(-1,8,-2, l, id); *dest++ = Vertex(-1,0,-2, l, id); *dest++ = Vertex(-1,0, 2, l, id);

		*dest++ = Vertex( 0,8, 2, f, id); *dest++ = Vertex(-1,8, 2, f, id); *dest++ = Vertex(-1,0, 2, f, id); *dest++ = Vertex( 0,0, 2, f, id);
		*dest++ = Vertex(-1,8,-2, b, id); *dest++ = Vertex( 0,8,-2, b, id); *dest++ = Vertex( 0,0,-2, b, id); *dest++ = Vertex(-1,0,-2, b, id);

		*dest++ = Vertex(-1,8,-2, u, id); *dest++ = Vertex(-1,8, 2, u, id); *dest++ = Vertex( 0,8, 2, u, id); *dest++ = Vertex( 0,8,-2, u, id);
		*dest++ = Vertex( 0,0,-2, d, id); *dest++ = Vertex( 0,0, 2, d, id); *dest++ = Vertex(-1,0, 2, d, id); *dest++ = Vertex(-1,0,-2, d, id);
	}

	if ((tileVB = renderer->addVertexBuffer(sizeof(vertices), STATIC, vertices)) == VB_NONE) return false;

	FormatDesc tileFormat[] = {
		{ 0, TYPE_VERTEX,   FORMAT_FLOAT, 3 },
		{ 0, TYPE_NORMAL,   FORMAT_FLOAT, 3 },
		{ 0, TYPE_TEXCOORD, FORMAT_FLOAT, 1 },
	};
	if ((tileVF = renderer->addVertexFormat(tileFormat, elementsOf(tileFormat))) == VF_NONE) return false;

	return true;
}

void App::unload(){

}

void App::drawFrame(){
	// Control the fading in and out
	if (time < 1){
		fade = time;
	} else if (attractionTile < tiles.getCount()){
		fade = 1;
	} else {
		static float first = true;
		static float fadeDown;

		if (first){
			fadeDown = 2;
			first = false;
		}

		fadeDown -= frameTime;
		if (fadeDown < 1){
			fade = fadeDown;

			if (fadeDown < 0){
				time = 0;
				resetCamera();
				arrangeTiles();
				first = true;
			}
		}
	}

	float t = time * 3.5f - 4.0f;

	// Step forward to the currently falling tile
	float c = 0;
	vec2 p(0, 0);
	while (attractionTile < tiles.getCount() && t - tiles[attractionTile].w > 0.0f){
		p += tiles[attractionTile].xy();
		attractionTile++;
		c++;
	}
	if (c) attractionPoint = p / c;


	if (!freeFly->isChecked()){
		float ft = min(frameTime, 0.05f);

		// Compute new camera position
		vec2 d = camPos.xz() - attractionPoint;
		float f = 25.0f / length(d) - 1.0f;
		vec2 dir = 1000.0f * ft * d * f;

		camPos.x += ft * dir.x;
		camPos.y = 12;
		camPos.z += ft * dir.y;


		// Compute new camera view angles
		float ad = wy - atan2f(d.x, -d.y);
		if (ad < -PI) ad += 2 * PI;
		if (ad >  PI) ad -= 2 * PI;
		ad = clamp(ad, -0.8f, 0.8f);

		float da = -3000 * ft * ad;

		wy += ft * da;
		if (wy < -PI) wy += 2 * PI;
		if (wy >  PI) wy -= 2 * PI;
	}




	mat4 projection = perspectiveMatrixX(1.5f, width, height, 0.1f, 2000);
	mat4 mvRotate = rotateXY(-wx, -wy);
	mat4 modelview = mvRotate * translate(-camPos);

	glMatrixMode(GL_PROJECTION);
	glLoadTransposeMatrixf(projection);

	glMatrixMode(GL_MODELVIEW);
	glLoadTransposeMatrixf(modelview);


	glClearColor(0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);



	// Draw the tiles
	renderer->reset();
	renderer->setRasterizerState(cullBack);
	renderer->setShader(tileShader);
	renderer->setVertexFormat(tileVF);
	renderer->setVertexBuffer(0, tileVB);
	renderer->setTexture("Noise", noise, trilinear);
	renderer->setShaderConstant3f("camPos", camPos);
	renderer->setShaderConstant1f("time", t);
	renderer->setShaderConstant1f("fade", fade);
	renderer->setShaderConstant1f("fallAngle", 1.22f);
	renderer->apply();

	int count = tiles.getCount() - 1;
	int tile = 0;
	while (count){
		int n = min(count, BATCH_INSTANCES);

		renderer->setShaderConstantArray4f("attribs", tiles.getArray() + tile, n);
		renderer->applyConstants();

		glDrawArrays(GL_QUADS, 0, n * 6 * 4);

		tile += n;
		count -= n;
	}

	// Last tile falls all the way to the ground
	renderer->setShaderConstantArray4f("attribs", tiles.getArray() + tiles.getCount() - 1, 1);
	renderer->setShaderConstant1f("fallAngle", PI / 2);
	renderer->applyConstants();

	glDrawArrays(GL_QUADS, 0, 6 * 4);




	// Draw the floor
	renderer->reset();
	renderer->setShader(floorShader);
	renderer->setTexture("Base", floorBase, trilinearAniso);
	renderer->setTexture("Bump", floorBump, trilinearAniso);
	renderer->setShaderConstant3f("camPos", camPos);
	renderer->setShaderConstant1f("fade", fade);
	renderer->apply();

	glBegin(GL_QUADS);
		glVertex3f(-1024, 0,  1024);
		glVertex3f( 1024, 0,  1024);
		glVertex3f( 1024, 0, -1024);
		glVertex3f(-1024, 0, -1024);
	glEnd();
}
